// ####################################################################################
// Includes
// ####################################################################################

#include "stdafx.h"
#include "internal.h"

using namespace GregorView::Internal;
using namespace GregorView::Managed;


// ####################################################################################
// Init/Fin
// ####################################################################################

CGregorViewFolderView::CGregorViewFolderView(CGregorViewShellFolder * pShellFolder, REFCLSID rclsid, HWND hWndOwner)
{
	::GetInstanceCounter().Increment();
	m_pShellFolder = pShellFolder;
	m_pShellFolder->AddRef();
	m_ClassId = rclsid;
	m_hWndOwner = hWndOwner;
	m_hWnd = NULL;
}

CGregorViewFolderView::~CGregorViewFolderView(void)
{
	::GetInstanceCounter().Decrement();
	if(m_pShellFolder != NULL){
		m_pShellFolder->Release();
	}
}


// ####################################################################################
// IUnknown
// ####################################################################################

STDMETHODIMP CGregorViewFolderView::QueryInterface(REFIID riid, void ** ppObj)
{
	*ppObj = NULL;
	if(riid == IID_IUnknown){
		*ppObj = (void*)(IShellView*)this; // (IShellView*) because we don't derive directly from IUnknown
		this->AddRef();
		return NOERROR;
	}else if(riid == IID_IOleWindow){
		::GetTracer().Trace("CGregorViewFolderView::QueryInterface riid <IID_IOleWindow>");
		*ppObj = (void*)(IShellView*)this; // (IShellView*) because we don't derive directly from IUnknown
		this->AddRef();
		return NOERROR;
	}else if(riid == IID_IShellView){
		::GetTracer().Trace("CGregorViewFolderView::QueryInterface riid <IID_IShellView>");
		*ppObj = (void*)(IShellView*)this;
		this->AddRef();
		return NOERROR;
	}else{
		return E_NOINTERFACE;
	}
}

STDMETHODIMP_(ULONG) CGregorViewFolderView::AddRef(void)
{
	m_RefCounter.Increment();
	return m_RefCounter.GetCount();
}

STDMETHODIMP_(ULONG) CGregorViewFolderView::Release(void)
{
	if(m_RefCounter.Decrement()){
		delete this;
		return 0;
	}else{
		return m_RefCounter.GetCount();
	}
}


// ####################################################################################
// IShellView
// ####################################################################################

// ... implement IShellView methods
STDMETHODIMP CGregorViewFolderView::GetWindow(HWND * pHwnd) 
{ 
	// IOleWindow::GetWindow
	::GetTracer().Trace("CGregorViewFolderView::GetWindow");
	return E_NOTIMPL; 
}
 
STDMETHODIMP CGregorViewFolderView::ContextSensitiveHelp(BOOL fEnterMode) 
{
	// IOleWindow::ContextSensitiveHelp
	::GetTracer().Trace("CGregorViewFolderView::ContextSensitiveHelp");
	return E_NOTIMPL; 
} 

STDMETHODIMP CGregorViewFolderView::TranslateAccelerator(LPMSG lpmsg) 
{ 
	::GetTracer().Trace("CGregorViewFolderView::TranslateAccelerator");
	return S_FALSE; // let the explorer process the key 
} 
 
 
STDMETHODIMP CGregorViewFolderView::EnableModeless(BOOL fEnable) 
{ 
	::GetTracer().Trace("CGregorViewFolderView::EnableModeless");
	return E_NOTIMPL; // currently not supposed to be implemented
} 
 
STDMETHODIMP CGregorViewFolderView::UIActivate(UINT uState) 
{ 
	::GetTracer().Trace("CGregorViewFolderView::UIActivate uState", (int)uState);
	return S_OK; 
} 

STDMETHODIMP CGregorViewFolderView::Refresh() 
{ 
	::GetTracer().Trace("CGregorViewFolderView::Refresh");
	return NOERROR; // ... give the implementation a chance 
} 
 
STDMETHODIMP CGregorViewFolderView::CreateViewWindow(
									IShellView  *lpPrevView, 
									LPCFOLDERSETTINGS lpfs,
									IShellBrowser  * psb, 
									RECT * prcView,
									HWND  *phWnd) 
{
	::GetTracer().TraceFormatted("CGregorViewFolderView::CreateViewWindow this <0x%x>", this);
	// reset out parameter
	*phWnd = NULL;
	// get parent window handle (IShellBrowser is an IOleWindow)
	HWND hWndParent = NULL;
	HRESULT hres = psb->GetWindow(&hWndParent);
	if(FAILED(hres)){
		return hres;
	}
	// get ex point
	CGregorViewExtensionPoint * pExPoint = GetExtensionPoints().GetOrCreate(m_ClassId);
	if(pExPoint == NULL){
		return E_UNEXPECTED;
	}
	// just use the root folder's view for now ...
	CFolderBase * pFolder = pExPoint->GetFolder(m_pShellFolder);
	if(pFolder == NULL){
		return E_UNEXPECTED;
	}
	IntPtr ptr = pFolder->GetViewWindow();
	if(ptr == IntPtr::Zero){
		return E_UNEXPECTED;
	}
	HWND hWnd = (HWND) ptr.ToPointer();
	::SetParent(hWnd, m_hWndOwner);
	::ShowWindow(hWnd, SW_SHOWDEFAULT);	
	// remember
	m_hWnd = hWnd;
	// set out parameter
	*phWnd = hWnd;
	// OK
	::GetTracer().TraceFormatted("CGregorViewFolderView::CreateViewWindow m_hWnd <0x%x>", m_hWnd);
	return S_OK; 
} 

STDMETHODIMP CGregorViewFolderView::DestroyViewWindow() 
{
	::GetTracer().TraceFormatted("CGregorViewFolderView::DestroyViewWindow this <0x%x> m_hWnd <0x%x>", this, m_hWnd);
	if(m_hWnd == NULL){
		return E_UNEXPECTED;
	}
	// hide, unparent
	ShowWindow(m_hWnd, SW_HIDE);
	SetParent(m_hWnd, NULL);	
	// give a chance to release
	CGregorViewExtensionPoint * pExPoint = GetExtensionPoints().GetOrCreate(m_ClassId);
	if(pExPoint == NULL){
		return E_UNEXPECTED;
	}
	CFolderBase * pFolder = pExPoint->GetFolder(m_pShellFolder);
	if(pFolder == NULL){
		return E_UNEXPECTED;
	}
	try{
		pFolder->ReleaseViewWindow(m_hWnd);
	}catch(Exception * pExc){
		(void)pExc;
	}
	m_hWnd = NULL;
	// OK
	::GetTracer().Trace("CGregorViewFolderView::CreateViewWindow success");
	return NOERROR; 
} 

STDMETHODIMP CGregorViewFolderView::GetCurrentInfo(LPFOLDERSETTINGS lpfs) 
{ 
	::GetTracer().Trace("CGregorViewFolderView::GetCurrentInfo");
	lpfs->ViewMode = FVM_DETAILS;
	lpfs->fFlags = 0;	
	return NOERROR; // ... give the implementation a chance 
} 

STDMETHODIMP CGregorViewFolderView::AddPropertySheetPages(DWORD dwReserved, 
                LPFNADDPROPSHEETPAGE lpfn, LPARAM lparam) 
{ 
	::GetTracer().Trace("CGregorViewFolderView::AddPropertySheetPages");
	return E_NOTIMPL; 
}
 
STDMETHODIMP CGregorViewFolderView::SaveViewState() 
{ 
	::GetTracer().Trace("CGregorViewFolderView::SaveViewState");
	return NOERROR; // ... give the implementation a chance 
} 
 
STDMETHODIMP CGregorViewFolderView::SelectItem(LPCITEMIDLIST pidlItem, UINT uFlags) 
{ 
	::GetTracer().Trace("CGregorViewFolderView::SelectItem");
	return E_NOTIMPL;
} 
 
STDMETHODIMP CGregorViewFolderView::GetItemObject(
											UINT uItem,
											REFIID riid, 
							                LPVOID *ppv)
{ 
	::GetTracer().Trace("CGregorViewFolderView::GetItemObject uItem", (int)uItem);
	return E_NOTIMPL; 
}