// ####################################################################################
// Includes
// ####################################################################################

#include "stdafx.h"
#include "internal.h"

using namespace GregorView::Internal;
using namespace GregorView::Managed;


// ####################################################################################
// Init/Fin
// ####################################################################################

CGregorViewShellFolder::CGregorViewShellFolder(REFCLSID rclsid)
{
	::GetInstanceCounter().Increment();
	// save the class ID
	m_ClassId = rclsid;
	// set absolute pidl to NULL (will be set in IPersistFolder::Initialize)
	m_pIDList = NULL;
}

CGregorViewShellFolder::~CGregorViewShellFolder(void)
{
	if(m_pIDList != NULL){
		::GetAllocator().Free((void*)m_pIDList);
	}
	::GetInstanceCounter().Decrement();
}


// ####################################################################################
// IUnknown
// ####################################################################################

STDMETHODIMP CGregorViewShellFolder::QueryInterface(REFIID riid, void ** ppObj)
{
	*ppObj = NULL;
	if(riid == IID_IUnknown){
		*ppObj = (void*)(IShellFolder*)this; // (IShellFolder*) because we don't derive directly from IUnknown
		this->AddRef();
		return NOERROR;
	}else if(riid == IID_IPersistFolder){
		::GetTracer().Trace("CGregorViewShellFolder::QueryInterface riid <IID_IPersistFolder>");
		*ppObj = (void*)(IPersistFolder*)this;
		this->AddRef();
		return NOERROR;
	}else if(riid == IID_IShellFolder){
		::GetTracer().Trace("CGregorViewShellFolder::QueryInterface riid <IID_IShellFolder>");
		*ppObj = (void*)(IShellFolder*)this;
		this->AddRef();
		return NOERROR;		
	}else{
		return E_NOINTERFACE;
	}
}

STDMETHODIMP_(ULONG) CGregorViewShellFolder::AddRef(void)
{
	m_RefCounter.Increment();
	return m_RefCounter.GetCount();
}

STDMETHODIMP_(ULONG) CGregorViewShellFolder::Release(void)
{
	if(m_RefCounter.Decrement()){
		delete this;
		return 0;
	}else{
		return m_RefCounter.GetCount();
	}
}


// ####################################################################################
// IShellFolder
// ####################################################################################

// ... implement IShellFolder methods

STDMETHODIMP CGregorViewShellFolder::CreateViewObject( 
										HWND hwndOwner,  
										REFIID riid, 
										LPVOID FAR* ppvObj) 
{
	::GetTracer().Trace("CGregorViewShellFolder::CreateViewObject");
	// check
	if(riid != IID_IShellView){
		return E_NOINTERFACE;
	}
	// create
	CGregorViewFolderView * pFolderView = new CGregorViewFolderView(this, m_ClassId, hwndOwner);
	if(pFolderView == NULL){
		return E_OUTOFMEMORY;
	}
	// set out parameter
	pFolderView->AddRef();
	*ppvObj = (void*)pFolderView;
	// OK
	return NOERROR;
} 
 
STDMETHODIMP CGregorViewShellFolder::EnumObjects( 
										HWND hwndOwner,  
										DWORD grfFlags, 
										LPENUMIDLIST * ppEnumIDList)
{ 
	::GetTracer().Trace("CGregorViewShellFolder::EnumObjects");
	// reset out parameter
	*ppEnumIDList = NULL;
	// get managed folder
	CFolderBase * pFolder = this->GetManagedFolder();
	if(pFolder == NULL){
		return E_UNEXPECTED;
	}
	// be sure to read in sub folders
	try{
		NFolderCollection * pSubFolders = pFolder->SubFolders;
		(void)pSubFolders;
	}catch(Exception * pExc){
		(void)pExc;
		return E_UNEXPECTED;
	}
	// create walker
	CGregorViewEnumerator * pWalker = new CGregorViewEnumerator(pFolder); // ... process flags
	if(pWalker == NULL){
		return E_OUTOFMEMORY;
	}
	// set out parameter
	pWalker->AddRef();
	*ppEnumIDList = (IEnumIDList*)pWalker;
	// OK
	return NOERROR;
}
 
STDMETHODIMP CGregorViewShellFolder::BindToObject( 
										LPCITEMIDLIST pidl,
										LPBC pbc,
										REFIID riid,
										LPVOID FAR* ppvObj)
{
	// reset out parameter
	*ppvObj = NULL;
	// we support only IShellFolder for now
	if(riid == IID_IShellFolder){
		::GetTracer().TraceFormatted("CGregorViewShellFolder::BindToObject riid <IID_IShellFolder>");
		// create folder
		CGregorViewShellFolder * pFolder = new CGregorViewShellFolder(m_ClassId);
		if(pFolder == NULL){
			return E_OUTOFMEMORY;
		}
		// save pidl (can be multi-level, but is always relative to this folder)
		pFolder->m_pIDList = ::ConcatIDLists(m_pIDList, pidl);
		// pidl info
		::GetTracer().TraceFormatted("CGregorViewShellFolder::BindToObject this <0x%x>", this);
		::GetTracer().TraceFormatted("CGregorViewShellFolder::BindToObject parent abs pidl %s", ::GetIDListInfoString(m_pIDList));
		::GetTracer().TraceFormatted("CGregorViewShellFolder::BindToObject child  rel pidl %s", ::GetIDListInfoString(pidl));
		::GetTracer().TraceFormatted("CGregorViewShellFolder::BindToObject child  abs pidl %s", ::GetIDListInfoString(pFolder->m_pIDList));
		// return
		pFolder->AddRef();
		*ppvObj = (void*)pFolder;
		return NOERROR;
	}else{
		return E_NOINTERFACE;
	}
}
 
STDMETHODIMP CGregorViewShellFolder::BindToStorage(
										LPCITEMIDLIST pidl,  
										LPBC pbc, 
										REFIID riid,  
   LPVOID FAR* ppvObj)
{
	::GetTracer().Trace("CGregorViewShellFolder::BindToStorage"); 
	return E_NOTIMPL;
} 

STDMETHODIMP CGregorViewShellFolder::CompareIDs(
										LPARAM lParam,
										LPCITEMIDLIST pidl1,
										LPCITEMIDLIST pidl2)
{
	::GetTracer().Trace("CGregorViewShellFolder::CompareIDs");
	// reinterpret some
	CGregorViewId * pID1 = CGregorViewId::FromItemID(pidl1);
	CGregorViewId * pID2 = CGregorViewId::FromItemID(pidl2);
	// ... ask managed DLL for a comparator object, and use that
	// delegate comparison to ID class
	short sResult = pID1->CompareTo(pID2);
	// return the result as an HRESULT
	HRESULT hres = MAKE_HRESULT(SEVERITY_SUCCESS, 0, sResult);
	return hres;
}

STDMETHODIMP CGregorViewShellFolder::ParseDisplayName(
										HWND hwndOwner,
										LPBC pbc,
										LPOLESTR lpszDisplayName,
										ULONG FAR* pchEaten,
										LPITEMIDLIST * ppidl,
										ULONG *pdwAttributes)
{ 
	::GetTracer().Trace("CGregorViewShellFolder::ParseDisplayName");
	// convert to ANSI
	char sz[MAX_PATH];
	if(0 == ::WideCharToMultiByte(CP_ACP, 0, lpszDisplayName, -1, sz, sizeof(sz), NULL, NULL)){
		return E_UNEXPECTED;
	}
	::GetTracer().TraceFormatted("CGregorViewShellFolder::ParseDisplayName path <%s>", sz);
	// validate display name (is relative to the folder called on,
	// can be multi-level, can be called on any of our folders)
	// ... support non-folders as well
	// get this shell folder's managed companion
	CFolderBase * pFolder = this->GetManagedFolder();
	if(pFolder == NULL){
		return E_UNEXPECTED;
	}
	// parse
	const ITEMIDLIST * pcidl = NULL;
	char * psz = strtok(sz, "\\");
	while(psz != NULL){
		// get sub folder at this level
		System::String * psSubDisplayName = CStringToSysString(psz);
		if(psSubDisplayName == NULL){
			return E_OUTOFMEMORY;
		}
		CFolderBase * pSubFolder = pFolder->SubFolders->get_Element(psSubDisplayName);
		if(pSubFolder == NULL){
			return E_INVALIDARG; // probably
		}
		// get attributes
		DWORD dwAttribs = (DWORD)pSubFolder->ShellAttributes;
		// create single-level item ID
		CGregorViewId * pID = new CGregorViewId(dwAttribs, psz);
		if(pID == NULL){
			return E_OUTOFMEMORY;
		}
		// concat?
		if(pcidl == NULL){
			pcidl = pID->AsItemID();
		}else{
			const ITEMIDLIST * pcidlNew = ::ConcatIDLists(pcidl, pID->AsItemID());
			if(pcidlNew == NULL){
				return E_OUTOFMEMORY;
			}
			::GetAllocator().Free((ITEMIDLIST*)pcidl);
			delete pID;
			pcidl = pcidlNew;
		}
		// eat away
		if(pchEaten != NULL){
			(*pchEaten) += (ULONG)strlen(psz);
		}
		// next single-level name
		psz = strtok(NULL, "\\");
		// next folder (pSubFolder is set)
		pFolder = pSubFolder;
	}
	*ppidl = const_cast<ITEMIDLIST*>(pcidl);
	// ... set attributes to nada if client can't be arsed to make another call
	if(pdwAttributes != NULL){
		 *pdwAttributes = 0;
	}
	// OK
	return NOERROR;
}

STDMETHODIMP CGregorViewShellFolder::GetDisplayNameOf( 
										LPCITEMIDLIST pidl,  
										DWORD dwFlags,  
										LPSTRRET lpName) 
{
	::GetTracer().Trace("CGregorViewShellFolder::GetDisplayNameOf");
	// so
	if((dwFlags & SHGDN_INFOLDER) == SHGDN_INFOLDER){
		::GetTracer().Trace("CGregorViewShellFolder::GetDisplayNameOf SHGDN_INFOLDER");
		// relative display name (under folder)
		CGregorViewId * pID = CGregorViewId::FromItemID(pidl);
		pID->GetName(lpName);
	}else{
		::GetTracer().Trace("CGregorViewShellFolder::GetDisplayNameOf SHGDN_NORMAL");
		// get extension point, connector
		CGregorViewExtensionPoint * pExPoint = GetExtensionPoints().GetOrCreate(m_ClassId);
		if(pExPoint == NULL){
			return E_UNEXPECTED;
		}
		CConnector * pConnector = pExPoint->GetConnector();
		if(pConnector == NULL){
			return E_UNEXPECTED;
		}		
		// ... what about virtual folder class IDs above the NSE root?
		// absolute display name (for editing, address bar, parsing)
		char sz[MAX_PATH];
		if((dwFlags & SHGDN_FORPARSING) == SHGDN_FORPARSING){
			strcpy(sz, "::");
			strcat(sz, pExPoint->GetClassIdAsString());
		}else{
			CFolderBase * pRootFolder = pConnector->GetRootFolder();
			if(pRootFolder == NULL){
				return E_UNEXPECTED;
			}
			const char * psz = SysStringToCString(pRootFolder->DisplayName);
			if(psz == NULL){
				return E_UNEXPECTED;
			}
			strcpy(sz, psz);
			FreeHGlobal(psz);
		}
		// concat names of parent folders (ID list must be relative to the NSE root folder)
		const ITEMIDLIST * pcidlThisUnderNseRoot = pExPoint->SkipNseRoot(m_pIDList);
		if(pcidlThisUnderNseRoot == NULL){
			return E_UNEXPECTED;
		}
		::ConcatPathOfNamesTo(pcidlThisUnderNseRoot, sz, sizeof(sz));
		// add names of items (ID list can be multi level if SHGDN_FORPARSING is set, but ID list is relative to this folder)
		::ConcatPathOfNamesTo(pidl, sz, sizeof(sz));
		// OK
		lpName->uType = STRRET_CSTR;
		strncpy(lpName->cStr, sz, MAX_PATH);
		lpName->cStr[MAX_PATH - 1] = '\0';
	}
	return NOERROR;
}

STDMETHODIMP CGregorViewShellFolder::SetNameOf( 
										HWND hwndOwner,  
										LPCITEMIDLIST pidl, 
										LPCOLESTR lpszName,  
										DWORD dwReserved, 
										LPITEMIDLIST FAR* ppidlOut) 
{ 
	::GetTracer().Trace("CGregorViewShellFolder::SetNameOf");
	return E_NOTIMPL;
}

STDMETHODIMP CGregorViewShellFolder::GetAttributesOf( 
										UINT cidl,  
										LPCITEMIDLIST FAR* apidl, 
										ULONG FAR* rgfInOut) 
{ 
	::GetTracer().Trace("CGregorViewShellFolder::GetAttributesOf");
	// walk the request IDs
	ULONG uCommonAttribs = 0xFFFFFFFF; 
	for(UINT u = 0; u < cidl; u++){
		// reinterpret
		CGregorViewId * pID = CGregorViewId::FromItemID(apidl[u]);
		// get attributes
		ULONG uAttribs = pID->GetAttribs();
		// filter out attribs not common to all IDs in the array
		uCommonAttribs = (uCommonAttribs & uAttribs);
	}
	// only one single ULONG is there, which contains only common flags
	*rgfInOut = uCommonAttribs;
	// OK
	return NOERROR;
}

STDMETHODIMP CGregorViewShellFolder::GetUIObjectOf(
										HWND hwndOwner,  
										UINT cidl,  
										LPCITEMIDLIST FAR* apidl,  
										REFIID riid,  
										UINT FAR* prgfInOut,  
										LPVOID FAR* ppvObj) 
{
	::GetTracer().Trace("CGregorViewShellFolder::GetUIObjectOf");
	// reset
	*ppvObj = NULL;
	// switch
	//if(riid == IID_IExtractIcon){ // cidl can only be 1
	//	::GetTracer().TraceFormatted("CGregorViewShellFolder::GetUIObjectOf riid <IID_IExtractIcon> cidl <%d>", cidl);
	//	CGregorViewIconExtractor * pExtractor = new CGregorViewIconExtractor();
	//	pExtractor->AddRef();
	//	*ppvObj = (IExtractIcon*)pExtractor;
	//	return NOERROR;
	//}else{
		return E_NOINTERFACE;
	//}
}


// ####################################################################################
// IPersistFolder
// ####################################################################################

STDMETHODIMP CGregorViewShellFolder::GetClassID(CLSID * pClassID)
{
	::GetTracer().Trace("CGregorViewShellFolder::GetClassID");
	*pClassID = m_ClassId;
	return S_OK;
}

STDMETHODIMP CGregorViewShellFolder::Initialize(const ITEMIDLIST * pidl)
{
	::GetTracer().TraceFormatted("CGregorViewShellFolder::Initialize this <0x%x> pidl %s", this, ::GetIDListInfoString(pidl));
	// get extension point
	CGregorViewExtensionPoint * pExPoint = GetExtensionPoints().GetOrCreate(m_ClassId);
	if(pExPoint == NULL){
		return E_UNEXPECTED;
	}
	// keep a clone of the ID list leading to the NSE's root folder
	pExPoint->SaveNseRoot(pidl);
	// keep a clone of the ID list
	m_pIDList = ::CloneIDList(pidl);
	::GetTracer().TraceFormatted("CGregorViewShellFolder::Initialize this <0x%x> m_pIDList %s", this, ::GetIDListInfoString(m_pIDList));
	// OK
	return NOERROR;
}


// ####################################################################################
// Other
// ####################################################################################

const ITEMIDLIST * CGregorViewShellFolder::GetIDList(void)
{
	return m_pIDList;
}

CFolderBase * CGregorViewShellFolder::GetManagedFolder(void)
{
	// get extension point
	CGregorViewExtensionPoint * pExPoint = GetExtensionPoints().GetOrCreate(m_ClassId);
	if(pExPoint == NULL){
		return NULL;
	}	
	// get
	CFolderBase * pFolder = pExPoint->GetFolder(this);
	// return
	return pFolder;
}
